/*****************************************************************************
  program name: lm_auto.c
  
  Program to do posterior IBD probs 
  The state (X) consists of segregation indicators -- now generalized so that 
  up to 8 gametes can be scored for IBD,
        
  X determines the IBD pattern among the specified descendant gametes.

  The data, Y, is the multilocus genotypes of observed individuals in the 
  pedigree.

  Modified from autozyg.c, April 1997,  so that sampling can be done on a
  meiosis by meiosis basis. (EAT)
  Modified to include general penetrance routine;          SCH/EAT Aug 1997
  Modified to score the Iceland states,                        EAT Oct 1997 
 
  Many improvements made May-July 1998, for GAW 11 analyses    EAT Jul 1998
     Now scores multiple gametes (up to 8) at multiple loci (up to 24) on 
     several component pedigrees (up to 5)

  Modified to allow use of the L-sampler,                       SCH Jul 1999 
  Modified to incorporate sex-specific maps, 
     to use L-sampler and M-sampler new library routines,       EAT Mar 2000
  Converted to MORGAN_V2.3                                      MAJ Jun 2000
  Modification made for mutiple scoring sets                    EAT Mar 2001
  Added AWG's sequential imputation set-up                      EAT Apr 2001
  Added parameter statements for multiple scoring sets &
     sampler setup                                              MAJ May 2001 
  Added mallocs and frees;  the number of gametes to be scored  
  is now between 2 and 10 (if sufficient memory is available)   MAJ Aug 2001

******************************************************************************/

#define PNAM  "lm_auto"

#include "autozyg.h"
#include "set_auto.c"
#include "alloc_auto.c"

int   main (int argc, char *argv[])
{
   int   i, j, k, l, ng, m, s;
   int   nloc;
   int ***y;                            /* full genotypic array */
   double **quant = NULL;
   int **dodo = NULL;
   int   t, c, h, err;
   int   jj;
   int   ngenes = 0, nfou;
   int   nsw, sx, *nns;
   int  *perm_sw, *perm_loc;            /* permutation arrays of switches 
                                         * and loci */
   int  *gaga;                          /* [g]          */
   int  *npt;                           /* [g+1]  no. of patterns for gametes */
   int **index;                         /* [g+1][cd+1]  */
   int **label;                         /* [g+1][pt+1]  */
   int **list;                          /* [g+1][pt+1]  */
   double ****nscore;                   /* [nc][ns][nl][pt] 
                                         *    ibd counts, ea component,
                                         *    ea score-set, ea locus, 
                                         *    ea pattern */
   int **count;                         /* [nc][nl]     */
   int ***recounts;                     /* [nc][nl][6] counts of recombinations
                                         */
   double *sss;                         /* [nl]         */
   double ss;
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   double **prob;

   int   tot_count = 0;                 /* score by component and locus */

   FILE *pedF, *oScorF, *oSeedF;
   FILE *ibdF = NULL;

   atexit (print_tail);
   PCleanUp = clean_auto;
   atexit (clean_morg);
   print_head (*argv);

   if (!proc_pars (argc, argv, PNAM))   /* process input parameter statements */
      exit (EXIT_FAILURE);

   if (!proc_auto ())                   /* check statements for completeness  */
      exit (EXIT_FAILURE);              /* and consistency  
                                         * and prepare controls for lm_auto   */

   if (!proc_seeds ())                  /* set up seed parameters             */
      exit (EXIT_FAILURE);

   if (!proc_fils ())                   /* set up pedigree file parameters    */
      exit (EXIT_FAILURE);

   setSeeds (SAMPL);                    /* initializes seeds for rans()       */

   if (!chk_pedfil ())                  /* checks pedigee file for errors     */
      give_up (ERR_qit);

   if (!mak_opedfil ())                 /* writes new pedigree file,          */
      give_up (ERR_qit);                /* if there are changes               */

   if (!(pedF = open_ped ()))           /* reopens pedigree file              */
      give_up (ERR_qit);

   alloc_ped ();
   input_ped (pedF, 0, dodo, quant);    /*input pedigree and trait data       */
   fclose (pedF);

   nloc = NLoci - 1;

   nfou = count_founders ();
   printf (" Number of founders is %d\n", nfou);

   MaxNsw = 2 * (pedsize - nfou);
   nns = malloc ((n_compnts + 1) * sizeof (int));
   alloc_chk (WHERE, nns);

   qdex = calloc (pedsize + 1, sizeof (int));
   alloc_chk (WHERE, qdex);
   trtphen = calloc (pedsize + 1, sizeof (int));
   alloc_chk (WHERE, trtphen);

   /* allocate space for genotypes y for our values of pedsize, NMarks */

   allocz_y (&y, nloc, pedsize);

   /* now set marker data on observed individuals */
   /* marker loci are labelled 1,..., nloc */

   set_markdat (y, nloc, n_compnts);

   /* set data for the trait */

   set_traitdat (y, qdex, trtphen, pedsize, n_compnts);

   allocate (pedsize, nloc);

   /* now set and index switches */

   nsw = list_switches (nloc, nns);
   printf (" Number of meiosis switches to sample is %d\n\n", nsw);
   fflush (stdout);

   /* set up permutation arrays for loci and switches  if Scan (not Step) */

   if (Scan)
   {
      perm_sw = malloc (nsw * sizeof (int));
      alloc_chk (WHERE, perm_sw);
      for (i = 0; i < nsw; ++i)
         perm_sw[i] = i;
      perm_loc = malloc ((nloc + 1) * sizeof (int));
      alloc_chk (WHERE, perm_loc);
      for (i = 0; i <= nloc; ++i)
         perm_loc[i] = i;
   }

   /* set up probands for gene scoring */

   set_probgam (n_compnts);

   /* set up recombination patterns and initial probs */

   /* Use L-sampler or sequential imputation for set_up and initial xx0, xx1 
      (maternal and paternal segregation indicators) */
   if (SmplSetup == LSMPL)              /* use L-sampler for set-up  */
      set_lsmpl (0, nloc, NAll, Qq, y);

   else                                 /* use sequential imputation */
      seq_imp (NSEQIMP, pedsize, nloc + 1, NAll, Qq, y, 0,
               (double *) 0, 1.0);      /* (AWG) */

   peel_trace = TRACE_LEVEL_0;          /* Turn on trace if level >0 */

   alloc_az_prob (&count, &recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                  &tot_log_prob_pen, &prob, &sss);

   /* set genes from initial switches */

   for (j = k = 0; j <= nloc; ++j)
   {
      ngenes = set_genes (j);

      /* Allocate memory for gen_pen() */

      if (loci[j].locus_type == TRAIT_LOCUS)
         alloc_gen_pen (ngenes, j);
      else if (!k++)
         alloc_seg_pen (ngenes, nloc + 1, y);

      for (c = 0; c < n_compnts; ++c)
         prob[c][j] = seg_pen2 (c, y, Qq[j], j, ngenes, &err);
   }
   tot_count = 0;

   /* gaga[], index[][], label[][], list[][], npt[] are malloced for the 
      number of gametes in the largest set */

   alloc_ibd (&gaga, &index, &label, &list, &npt, &nscore);

   /* count total number of patterns ----
      modified EAT, March 2001, now will do only full set of patterns
      for each number of gametes  */

   /* probability now computed separately by component  */

   for (c = 0; c < n_compnts; ++c)
   {
      for (h = 0; h <= nloc; ++h)
      {
         tot_log_prob_seg[c][h] = 0.0;
         tot_log_prob_pen[c][h] = 0.0;
      }
      for (j = 0; j <= nloc; ++j)
      {
         count[c][j] = 0;
         for (jj = 0; jj < 6; ++jj)
            recounts[c][j][jj] = 0;
         for (s = 0; s < NScSet[c]; ++s)
            for (k = 0; k <= PTgg[NPrGam[c][s]]; ++k)
               nscore[c][s][j][k] = 0.0;
      }
   }
   tot_count = 0;

   printf ("  Number of pattern sets to be scored:\n");
   npt[j] = 0;
   for (j = 1; j <= MxGam; ++j)
   {
      npt[j] = index_ibd (j, 0, 0, index[j], label[j], list[j], ibdF);

      k = 0;
      for (c = 0; c < n_compnts; ++c)
         for (s = 0; s < NScSet[c]; ++s)
            if (j == NPrGam[c][s])
               k = 1;
      if (k)
         printf ("    on %2d gametes:%7d \n", j, npt[j]);
   }
   printf ("\n\n");
   fflush (stdout);

   alloc_logprob ();

   alloc_msmpl ();

   /* open output scores file, if requested */

   if (FWIt)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   /* now test switches and do MC sampling */

   for (t = 0; t < NIt; ++t)
   {
      if (rans () < LSamplProb)
      {                                 /* ***** Do L-sample *****  */
         if (!((t + 1) % WIt))
         {
            printf (" Starting L-sampler iteration %d\n", t + 1);
            fflush (stdout);
         }
         if (Scan)                      /* ****   scan   **** */
         {
            permute (perm_loc, nloc + 1);  /* permute loci for L-sampler */
            for (k = 0; k <= nloc; ++k)
            {                           /* do l-sampling for ea. locus */
               jj = perm_loc[k];
               do_lsmpl (nloc, jj, prob, y, ngenes);
            }
            for (jj = 0; jj <= nloc; ++jj)  /* score ea. gamete set for */
            {                           /* ea. locus */
               j = revchro[0][jj];
               for (c = 0; c < n_compnts; ++c)
               {
                  ++count[c][j];
                  for (s = 0; s < NScSet[c]; ++s)
                     score_genes (jj, NPrGam[c][s], gaga, index[NPrGam[c][s]],
                                  list[NPrGam[c][s]], Proband[c][s],
                                  ProbandGamete[c][s], nscore[c][s][j]);
               }
            }
         }
         else
         {                              /* ****   step   **** */

            jj = intran (0, nloc);      /* choose a random locus
                                         * do l-sampling for this locus */
            do_lsmpl (nloc, jj, prob, y, ngenes);

            j = revchro[0][jj];         /* score ea. gamete set for */
            for (c = 0; c < n_compnts; ++c)  /* for this locus */
            {
               ++count[c][j];
               for (s = 0; s < NScSet[c]; ++s)
                  score_genes (jj, NPrGam[c][s], gaga, index[NPrGam[c][s]],
                               list[NPrGam[c][s]], Proband[c][s],
                               ProbandGamete[c][s], nscore[c][s][j]);
            }
         }
      }
      else
      {                                 /* ***** Do M-sample *****  */
         if (!((t + 1) % WIt))
         {
            printf (" Starting M-sampler iteration %d\n", t + 1);
            fflush (stdout);
         }
         if (Scan)                      /* ****   scan   **** */
         {
            permute (perm_sw, nsw);     /* permute switches for M-sampler */
            for (k = 0; k < nsw; ++k)
            {                           /* for ea. switch, get meiosis */
               l = perm_sw[k] + 1;      /* and component */
               i = (meiosis[l].ind).index;
               sx = meiosis[l].sex;
               c = nghd[i].component - 1;
               /*                        * do m-sampling for the meiosis */
               do_msmpl (nloc, c, i, sx, prob, y, ngenes);
            }
            for (j = nloc; j >= 0; --j) /* score ea. gamete set for */
            {                           /* ea. locus */
               jj = chromo[0][j];
               for (c = 0; c < n_compnts; ++c)
               {
                  ++count[c][j];
                  for (s = 0; s < NScSet[c]; ++s)
                     score_genes (jj, NPrGam[c][s], gaga, index[NPrGam[c][s]],
                                  list[NPrGam[c][s]], Proband[c][s],
                                  ProbandGamete[c][s], nscore[c][s][j]);
               }
            }
         }
         else
         {                              /* ****   step   **** */

            l = intran (0, nsw - 1) + 1;  /* choose a random switch */
            i = (meiosis[l].ind).index;
            sx = meiosis[l].sex;        /* get meiosis and component */
            c = nghd[i].component - 1;
            /*                           * do m-sampling for this meiosis */
            do_msmpl (nloc, c, i, sx, prob, y, ngenes);

            for (j = nloc; j >= 0; --j) /* score this component's */
            {                           /* gamete sets for ea. locus */
               jj = chromo[0][j];
               ++count[c][j];
               for (s = 0; s < NScSet[c]; ++s)
               {
                  score_genes (jj, NPrGam[c][s], gaga, index[NPrGam[c][s]],
                               list[NPrGam[c][s]], Proband[c][s],
                               ProbandGamete[c][s], nscore[c][s][j]);
               }
            }
         }
      }
      if (!((t + 1) % ScIt))
      {                                 /* score tot log-probs & recombs */
         for (c = 0; c < n_compnts; ++c)
         {
            full_seg_prob (nloc, c, recounts, temp_log_prob_seg);
            for (j = 0; j < nloc; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[c][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
            }
            tot_log_prob_pen[c][nloc] += log (prob[c][nloc]);
         }
         ++tot_count;
      }
      if (FWIt && !((t + 1) % FWIt))    /* tests for output to scores fil */
         for (c = 0; c < n_compnts; ++c)
         {                              /* outputs by component           */
            fprintf (oScorF, "%d ", c + 1);
            for (jj = 0; jj <= nloc; ++jj)
            {
               j = revchro[0][jj];
               fprintf (oScorF, "%d ", count[c][j]);  /* (counts, by name)    */
            }
            fprintf (oScorF, "\n");
            for (s = 0; s < NScSet[c]; ++s)
            {                           /* then by proband gamete set     */
               ng = NPrGam[c][s];
               for (k = 0; k < npt[ng]; ++k)
               {                        /* then by label and pattern      */
                  fprintf (oScorF, "%d ", s + 1);
                  invert_label (label[ng][k], ng, gaga);
                  fprintf (oScorF, "%d  ", label[ng][k]);
                  for (i = 0; i < ng; ++i)
                     fprintf (oScorF, "%d ", gaga[i]);
                  for (jj = 0; jj <= nloc; ++jj)  /* then by name             */
                  {
                     j = revchro[0][jj];
                     fprintf (oScorF, " %e", nscore[c][s][j][list[ng][k]]);
                  }
                  fprintf (oScorF, "\n");
               }
            }
         }
   }
   printf ("\n");

   for (c = 0; c < n_compnts; ++c)
   {
      printf ("\n Summary for component %d:\n\n", c + 1);

      printf ("    Average total log-probability of segregations:\n");
      printf ("       first locus (constant) %8.3f\n       by map interval:",
              -nns[c + 1] * log (2.0));

      for (j = 0; j < nloc; ++j)
      {
         if (!(j % 8))
            printf ("\n       ");
         printf (" %8.3f", tot_log_prob_seg[c][j] / tot_count);
      }
      printf ("\n\n");

      printf ("    Average total log-probability of penetrances,"
              " by locus:");
      for (j = 0; j <= nloc; ++j)
      {
         if (!(j % 8))
            printf ("\n      ");
         if (chromo[0][j] == 0)
            printf ("  trait  ");
         else
            printf (" mark%2d  ", chromo[0][j]);
      }
      for (j = 0; j <= nloc; ++j)
      {
         if (!(j % 8))
            printf ("\n    ");
         printf (" %8.3f", tot_log_prob_pen[c][chromo[0][j]] / tot_count);
      }
      printf ("\n\n");

      /* print out recombination counts  */
      printf ("    Recombination counts for map intervals:\n");
      printf ("       intv   M: total    recomb   non-rec "
              "  F: total    recomb   non-rec\n");
      for (j = 0; j < nloc; ++j)
      {
         printf ("       %4d ", j + 1);
         for (i = 0; i < 6; ++i)
         {
            if (i == 3)
               printf (" ");
            printf ("%10d", recounts[c][j][i]);
         }
         printf ("\n");
      }
      printf ("\n");
      fflush (stdout);

      for (s = 0, m = 4; s < NScSet[c]; ++s)
         if (m < NPrGam[c][s])
            m = NPrGam[c][s];

      for (s = 0; s < NScSet[c]; ++s)
      {
         ng = NPrGam[c][s];

         printf ("    Probabilities of ibd patterns for proband gamete set %d:"
                 ,s + 1);

         for (j = 0; j <= nloc; ++j)    /* format headings */
         {
            if (!(j % 8))
            {
               printf ("\n      ");
               for (i = 4; i < m; ++i)
                  printf ("  ");
               if (j == 0)
                  printf (" pattern  ");
               else
                  printf ("          ");
            }
            if (chromo[0][j] == 0)
               printf (" trait ");
            else
               printf ("mark%2d ", chromo[0][j]);
            if (j == nloc)
               printf (" label");
         }
         for (k = 0; k < npt[ng]; ++k)  /* format data */
         {
            ss = 0.0;
            for (j = 0; j <= nloc; ++j)
            {
               sss[j] = nscore[c][s][j][list[ng][k]] / count[c][j];
               ss += sss[j];
            }
            if (ss > 0.001)
            {
               invert_label (label[ng][k], ng, gaga);
               for (j = 0; j <= nloc; ++j)
               {
                  if (!(j % 8))
                  {
                     printf ("\n       ");
                     for (i = 0; i < m - ng; ++i)
                        printf ("  ");
                     if (j == 0)
                        for (i = 0; i < ng; ++i)
                           printf ("%d ", gaga[i]);
                     else
                        for (i = 0; i < ng; ++i)
                           printf ("  ");
                  }
                  printf ("%s", dub2nlz (sss[j], 7, 4));
                  if (j == nloc)
                     printf ("     %2d", label[ng][k]);
               }
            }
         }
         printf ("\n\n");
      }
   }
   printf ("\n");                       /* ****  now we are done **** */
   fflush (stdout);

   /* Free up allocated memory */

   free_y (&y, nloc);
   free_ibd (&gaga, &index, &label, &list, &npt, &nscore);
   free_az_prob (&count, &recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen, &prob, &sss);
   locus_peel_free ();
   free_allocate_loci ();
   free_gen_pen ();
   free_seg_pen (nloc + 1);
   free_logprob ();
   free_msmpl ();

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (nns)
   {
      free (nns);
      nns = NULL;
   }
   if (qdex)
   {
      free (qdex);
      qdex = NULL;
   }
   if (trtphen)
   {
      free (trtphen);
      trtphen = NULL;
   }

   exit (EXIT_SUCCESS);
}
